@@ -14,9 +14,9 @@ use std::collections::HashMap;
 use std::future::Future;
 
 use macro_rules_attribute::apply;
-use perspective_client::{assert_table_api, assert_view_api, Session};
 #[cfg(doc)]
-use perspective_client::{config::ViewConfigUpdate, Schema, TableInitOptions, UpdateOptions};
+use perspective_client::{Schema, TableInitOptions, UpdateOptions, config::ViewConfigUpdate};
+use perspective_client::{Session, assert_table_api, assert_view_api};
 use pyo3::exceptions::PyTypeError;
 use pyo3::marker::Ungil;
 use pyo3::prelude::*;
@@ -142,6 +142,20 @@ impl Client {
         self.0.get_hosted_table_names().py_block_on(py)
     }
 
+    #[apply(inherit_doc)]
+    #[inherit_doc = "client/on_hosted_tables_update.md"]
+    pub fn on_hosted_tables_update(&self, py: Python<'_>, callback: Py<PyAny>) -> PyResult<u32> {
+        self.0.on_hosted_tables_update(callback).py_block_on(py)
+    }
+
+    #[apply(inherit_doc)]
+    #[inherit_doc = "client/remove_hosted_tables_update.md"]
+    pub fn remove_hosted_tables_update(&self, py: Python<'_>, callback_id: u32) -> PyResult<()> {
+        self.0
+            .remove_hosted_tables_update(callback_id)
+            .py_block_on(py)
+    }
+
     #[apply(inherit_doc)]
     #[inherit_doc = "client/set_loop_callback.md"]
     pub fn set_loop_callback(&self, py: Python<'_>, loop_cb: Py<PyAny>) -> PyResult<()> {
